SOFTWARE_VERSION_NUMBER = "2.1.4"
YEAR_CONSTANT = 2023
# STELLA-Q software
# Science and Technology Education for Land/ Life Assessment
# Paul Mirel

HUGE_VALUE = False #99999 #False #654321 # sustitute for all irradiance values, to stress the available processor memory.
# change HUGE_VALUE to False to allow ordinary measurements of irradiance
MAX_VALUE = 99999 #cap irradiance values at 100,000 - 1 to prevent memory allocation crash

import gc #garbage collection, RAM management
gc.collect()
start_mem_free_kB = gc.mem_free()/1000
print("start memory free {0:.2f} kB".format( start_mem_free_kB ))

import time
start_time = time.monotonic()
# load system libraries
import os
import board
import microcontroller
import digitalio
from analogio import AnalogIn
import neopixel
import sdioio
import storage
import busio
import terminalio
import ulab
sys_lib_time = time.monotonic()
#print( "time needed to load system libraries == {:.1f} s".format( sys_lib_time - start_time ))
gc.collect()

# load device libraries
import adafruit_pcf8523                             # real time clock
import adafruit_sht4x                               # air temp and humidity
import adafruit_tca9548a                            # i2c mux
from adafruit_as726x import AS726x_I2C              # spectral sensors
from adafruit_neokey.neokey1x4 import NeoKey1x4     # neokeyboard
from adafruit_ht16k33 import segments               # 7 segment display driver
#import adafruit_mlx90614                            # thermal infrared optional
#device_lib_time = time.monotonic()
#print( "time needed to load all libraries == {:.1f} s".format( device_lib_time - start_time ))
#print( "memory free after load libraries {:.1f} kB".format( gc.mem_free()/1000 ))
gc.collect()
#print( "memory free after garbage collection {:.1f} kB".format( gc.mem_free()/1000 ))

from stella_q_utility import set_date_time
gc.collect()

VIS_BANDS =         ( 450, 500, 550, 570, 600, 650 ) # from amd as7262 datasheet
VIS_BAND_PREFIXES = ( "V", "B",  "G", "Y", "O", "R" )
NIR_BANDS =         ( 610, 680, 730, 760, 810, 860 ) # from amd as7263 datasheet
AH_ON =     True       #   for active high discrete devices
AH_OFF =    False      #   for active high discrete devices
AL_ON =     False      #   for active low discrete devices
AL_OFF =    True       #   for active low discrete devices
NEO_RED =       ( 255, 0, 0 )
NEO_OFF =       ( 0, 0, 0 )

DATA_FILE = "/sd/data.csv"

def main():
    surface_temperature_C = 0
    onboard_neopixel = initialize_neopixel( board.NEOPIXEL, 1 )
    set_neopixel( onboard_neopixel, NEO_RED )
    time.sleep( 0.1 )
    set_neopixel( onboard_neopixel, NEO_OFF )

    # initialize sd card file storage
    try:
        sdcard = sdioio.SDCard( clock=board.SDIO_CLOCK, command=board.SDIO_COMMAND, data=board.SDIO_DATA, frequency=25000000 )
        vfs = storage.VfsFat(sdcard)
    except( OSError, ValueError ):
        print( "No SD card found, or card is full" )
        vfs = False
    if vfs:
        storage.mount(vfs, "/sd")
        initialize_data_file()
    else:
        set_neopixel( onboard_neopixel, NEO_RED )
        print( "(future print to instrument display): SD card initialization failed. No SD card or card full." )
    #print( "write header == {}".format( write_header ))

    gc.collect()
    #print( "memory free after initialize virtual file system {:.1f} kB".format( gc.mem_free()/1000 ))


    LOW_BATTERY_VOLTAGE = 3.1

    RECORD =    0
    HOLD =      1
    LIVE =      2
    AVERAGE =   3

    SHOW_UID =              0
    SHOW_TIME =             1
    SHOW_BATCH_NUMBER =     2
    SHOW_SOFTWARE_VERSION = 3


    HEX_OFF =       0x000000
    HEX_ORANGE =    0xFF5500
    HEX_YELLOW =    0xFFFF00
    HEX_GREEN =     0x00FF00
    HEX_BLUE =      0x0000FF

    MUX_channel_NIR = 0
    MUX_channel_VIS = 1
    display_modified_address = 0x77

    power_indicator_LED = initialize_discrete_LED( board.LED )
    power_indicator_LED.value = AH_ON
    operational = True
    # initialize bus
    try:
        i2c_bus = initialize_i2c_bus( board.SCL, board.SDA )
    except:
        i2c_bus = False
        operational = False

    if operational:
        # initialize board devices
        UID = int.from_bytes(microcontroller.cpu.uid, "big") % 10000
        print( "unique identifier (UID) == {0}".format( UID ))
        print( "STELLA software version == {}".format( SOFTWARE_VERSION_NUMBER ))
        main_battery_monitor = initialize_AnalogIn( board.VOLTAGE_MONITOR )
        main_battery_voltage = check_battery_voltage( main_battery_monitor )
        #print( "main_battery_voltage == {}".format( main_battery_voltage ))

        gc.collect()
        #print( "memory free after initialize board devices {:.1f} kB".format( gc.mem_free()/1000 ))

    if operational:
        # initialize hardware devices
        rtc_group = initialize_real_time_clock( i2c_bus )
        rtc = rtc_group[ 0 ]
        rtc_battery_OK = rtc_group[ 1 ]
        days = ( "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" )
        # wday =    0         1         2           3            4          5          6
        if rtc:
            timestamp = rtc.datetime
            if timestamp.tm_year < 2023:
                timestamp = time.struct_time(( 2023,  01,   01,   12,  00,  00,   0,   -1,    -1 ))
                rtc.datetime = timestamp
        else:
            timestamp = time.struct_time(( 0000,  00,   00,   00,  00,  00,   0,   -1,    -1 ))



    if operational:
        batch_number = update_batch( timestamp )
        ( desired_sample_interval_s, averaging_number_of_samples,
            averaging_count_down, count_down_step_s ) = check_STELLA_config()

        # initialize sensors
        #TIR_sensor = initialize_TIR_sensor( i2c_bus )
        temp_hum_SHT40 = initialize_SHT40( i2c_bus )
        try:
            bus_mux = initialize_tca_i2c_multiplexer( i2c_bus )
            NIR_spectral_sensor = initialize_spectral_sensor( bus_mux, MUX_channel_NIR )
            VIS_spectral_sensor = initialize_spectral_sensor( bus_mux, MUX_channel_VIS )
        except OSError:
            bus_mux = False
            NIR_spectral_sensor = False
            VIS_spectral_sensor = False
            print( "spectral sensors not found" )
        gc.collect()
        #print( "memory free after initialize sensors {:.1f} kB".format( gc.mem_free()/1000 ))

        # set instrument mode
        instrument_mode = LIVE
        last_instrument_mode = HOLD

        four_digit_display = initialize_four_digit_display( i2c_bus, display_modified_address )
        if four_digit_display:
            four_digit_display.print( UID )
        neokey1x4 = initialize_neokey1x4( i2c_bus )
        if neokey1x4:
            neokey1x4.pixels[0] = HEX_ORANGE    # orange key changes display modes
            neokey1x4.pixels[1] = HEX_YELLOW    # yellow key takes one sample, or toggles set_check in date/time mode
            neokey1x4.pixels[2] = HEX_GREEN     # green key takes multiple samples and averages them, or adds to value in date/time mode
            neokey1x4.pixels[3] = HEX_BLUE      # blue toggles between continuous record and pause, or subtracts from value in date/time mo
        gc.collect()
        #print( "memory free after initialize keys and display {:.1f} kB".format( gc.mem_free()/1000 ))
    # set modes and controls
    display_mode = SHOW_UID
    number_of_display_modes = 4
    last_display_mode = 10 # startup
    orange_key_pressed =        False
    last_orange_key_pressed =   False
    blue_key_pressed =          False
    last_blue_key_pressed =     False
    yellow_key_pressed =        False
    last_yellow_key_pressed =   False
    green_key_pressed =         False
    last_green_key_pressed =    False
    try:
        while operational:
            gc.collect()
            last_sample_time_s = time.monotonic()
            if rtc:
                timestamp = rtc.datetime
            else:
                timestamp = time.struct_time(( 0000,  00,   00,   00,  00,  00,   0,   -1,    -1 ))
            iso8601_utc = "{:04}{:02}{:02}T{:02}{:02}{:02}Z".format( timestamp.tm_year,
                timestamp.tm_mon, timestamp.tm_mday, timestamp.tm_hour, timestamp.tm_min,
                timestamp.tm_sec
                )
            weekday = timestamp_to_weekday( timestamp )
            decimal_hour = timestamp_to_decimal_hour( timestamp )
            mem_free_kB = gc.mem_free()/1000
            print( "\nmemory free == {} kB, {:.1f} %".format( mem_free_kB, 100 * mem_free_kB/start_mem_free_kB ))

            if four_digit_display:
                if display_mode == SHOW_UID and last_display_mode != SHOW_UID:
                    four_digit_display.fill(0)
                    four_digit_display.print( UID )
                    last_display_mode = SHOW_UID
                elif display_mode == SHOW_TIME:
                    four_digit_display.fill(0)
                    four_digit_display.print( "{:02}:{:02}".format( timestamp.tm_hour, timestamp.tm_min ))
                    last_display_mode = SHOW_TIME
                elif display_mode == SHOW_BATCH_NUMBER:
                    four_digit_display.fill(0)
                    four_digit_display.print( batch_number )
                    four_digit_display[0] = ( "B" )
                    last_display_mode = SHOW_BATCH_NUMBER
                elif display_mode == SHOW_SOFTWARE_VERSION:
                    four_digit_display.fill(0)
                    four_digit_display.print( SOFTWARE_VERSION_NUMBER )
                    four_digit_display.set_digit_raw(0, 0x1C)
                    last_display_mode = SHOW_SOFTWARE_VERSION

            exit_wait = 0

            if instrument_mode == RECORD:
                #print( "instrument mode == RECORD" )
                if last_instrument_mode != RECORD:
                    batch_number = update_batch( timestamp )
                    exit_wait = 1
                    display_mode = SHOW_BATCH_NUMBER
                if VIS_spectral_sensor:
                    visible_irradiance_uW_per_cm_squared = read_spectrum( VIS_spectral_sensor )
                else:
                    visible_irradiance_uW_per_cm_squared = (0,0,0,0,0,0)
                if NIR_spectral_sensor:
                    nir_irradiance_uW_per_cm_squared = read_spectrum( NIR_spectral_sensor )
                else:
                    nir_irradiance_uW_per_cm_squared = (0,0,0,0,0,0)
                #surface_temperature_C = read_TIR_sensor( TIR_sensor )
                air_temperature_C, relative_humidity_percent = read_environmental_sensors( temp_hum_SHT40 )
                main_battery_voltage = check_battery_voltage( main_battery_monitor )
                write_line_success = write_data_to_file( UID, timestamp, decimal_hour, batch_number,
                    visible_irradiance_uW_per_cm_squared, nir_irradiance_uW_per_cm_squared,
                    surface_temperature_C, air_temperature_C, relative_humidity_percent,
                    main_battery_voltage, onboard_neopixel )
                mirror_data_over_usb(
                    UID, timestamp, decimal_hour, batch_number,
                    visible_irradiance_uW_per_cm_squared, nir_irradiance_uW_per_cm_squared,
                    surface_temperature_C, air_temperature_C, relative_humidity_percent)
                last_instrument_mode = RECORD

            if instrument_mode == HOLD:
                #print( "instrument mode == HOLD" )
                if last_instrument_mode != HOLD:
                    batch_number = update_batch( timestamp )
                    exit_wait = 1
                    display_mode = SHOW_BATCH_NUMBER
                    if VIS_spectral_sensor:
                        visible_irradiance_uW_per_cm_squared = read_spectrum( VIS_spectral_sensor )
                    else:
                        visible_irradiance_uW_per_cm_squared = (0,0,0,0,0,0)
                    if NIR_spectral_sensor:
                        nir_irradiance_uW_per_cm_squared = read_spectrum( NIR_spectral_sensor )
                    else:
                        nir_irradiance_uW_per_cm_squared = (0,0,0,0,0,0)
                    #surface_temperature_C = read_TIR_sensor( TIR_sensor )
                    air_temperature_C, relative_humidity_percent = read_environmental_sensors( temp_hum_SHT40 )
                    main_battery_voltage = check_battery_voltage( main_battery_monitor )
                    write_line_success = write_data_to_file( UID, timestamp, decimal_hour, batch_number,
                        visible_irradiance_uW_per_cm_squared, nir_irradiance_uW_per_cm_squared,
                        surface_temperature_C, air_temperature_C, relative_humidity_percent,
                        main_battery_voltage, onboard_neopixel )
                    mirror_data_over_usb(
                        UID, timestamp, decimal_hour, batch_number,
                        visible_irradiance_uW_per_cm_squared, nir_irradiance_uW_per_cm_squared,
                        surface_temperature_C, air_temperature_C, relative_humidity_percent)
                last_instrument_mode = HOLD

            if instrument_mode == LIVE:
                #print( "instrument mode == LIVE" )
                if VIS_spectral_sensor:
                    visible_irradiance_uW_per_cm_squared = read_spectrum( VIS_spectral_sensor )
                else:
                    visible_irradiance_uW_per_cm_squared = (0,0,0,0,0,0)
                if NIR_spectral_sensor:
                    nir_irradiance_uW_per_cm_squared = read_spectrum( NIR_spectral_sensor )
                else:
                    nir_irradiance_uW_per_cm_squared = (0,0,0,0,0,0)
                # surface_temperature_C = 0 #read_TIR_sensor( TIR_sensor )
                air_temperature_C, relative_humidity_percent = read_environmental_sensors( temp_hum_SHT40 )
                main_battery_voltage = check_battery_voltage( main_battery_monitor )
                mirror_data_over_usb(
                    UID, timestamp, decimal_hour, batch_number,
                    visible_irradiance_uW_per_cm_squared, nir_irradiance_uW_per_cm_squared,
                    surface_temperature_C, air_temperature_C, relative_humidity_percent)
                last_instrument_mode = LIVE

            if instrument_mode == AVERAGE and four_digit_display:
                if vfs:
                    if neokey1x4:
                        neokey1x4.pixels[2] = 0x00FF00
                    four_digit_display.fill( 0 )
                    #print( "instrument mode == AVERAGE" )
                    batch_number = update_batch( timestamp )
                    exit_wait = 1
                    count_down_index = averaging_count_down
                    #surface_temperature_C = read_TIR_sensor( TIR_sensor )
                    air_temperature_C, relative_humidity_percent = read_environmental_sensors( temp_hum_SHT40 )
                    main_battery_voltage = check_battery_voltage( main_battery_monitor )
                    VIS_sensor_temperature, NIR_sensor_temperature = get_spectral_sensor_temperatures( VIS_spectral_sensor, NIR_spectral_sensor )
                    header_values = ( UID, batch_number, iso8601_utc, surface_temperature_C, air_temperature_C, relative_humidity_percent, VIS_sensor_temperature, NIR_sensor_temperature )
                    sampling_filename = create_sampling_file( UID, batch_number, timestamp, decimal_hour, header_values )
                    while count_down_index > 0:
                        print( "count down remaining = {}".format( count_down_index ))
                        four_digit_display.fill( 0 )
                        four_digit_display[1] = "C"
                        four_digit_display[2] ="D"
                        four_digit_display.print( count_down_index )
                        count_down_index = count_down_index - 1
                        time.sleep( count_down_step_s )
                    sampling_index = averaging_number_of_samples
                    sample_average = [ 0,0,0,0,0,0,0,0,0,0,0,0 ]
                    while sampling_index > 0:
                        print( "sample number = {}".format( sampling_index ))
                        if VIS_spectral_sensor:
                            visible_irradiance_uW_per_cm_squared = read_spectrum( VIS_spectral_sensor )
                        else:
                            visible_irradiance_uW_per_cm_squared = (0,0,0,0,0,0)
                        if NIR_spectral_sensor:
                            nir_irradiance_uW_per_cm_squared = read_spectrum( NIR_spectral_sensor )
                        else:
                            nir_irradiance_uW_per_cm_squared = (0,0,0,0,0,0)
                        write_sampling_line_to_file( UID, batch_number, iso8601_utc, onboard_neopixel, sampling_filename, sampling_index, visible_irradiance_uW_per_cm_squared, nir_irradiance_uW_per_cm_squared )
                        spectral_samples = visible_irradiance_uW_per_cm_squared + nir_irradiance_uW_per_cm_squared
                        #print( spectral_samples )
                        for i,v in enumerate( spectral_samples ):
                            sample_average[ i ] = sample_average[ i ] + v
                        four_digit_display.fill( 0 )
                        four_digit_display.print( sampling_index )
                        sampling_index = sampling_index - 1
                    for i,v in enumerate( sample_average ):
                        sample_average[ i ] = round( sample_average[ i ]/ averaging_number_of_samples, 1 )
                    #print( sample_average )
                    write_tail_of_sampling_file( UID, batch_number, iso8601_utc,
                        onboard_neopixel, sampling_filename, averaging_number_of_samples,
                        sample_average, VIS_sensor_temperature, NIR_sensor_temperature, surface_temperature_C,
                        air_temperature_C, relative_humidity_percent
                        )
                    four_digit_display.print( "STDV" )
                    stdv_successs = calculate_and_write_standard_deviations( UID, batch_number, sampling_filename )
                    if stdv_successs:
                        four_digit_display.print( "DONE" )
                    else:
                        four_digit_display.print( "NODV" )
                    averaged_visible_irradiance_uW_per_cm_squared = [0,0,0,0,0,0]
                    averaged_nir_irradiance_uW_per_cm_squared = [0,0,0,0,0,0]
                    for i,v in enumerate( sample_average ):
                        if i < 6:
                            averaged_visible_irradiance_uW_per_cm_squared[ i ] = sample_average[ i ]
                        else:
                            averaged_nir_irradiance_uW_per_cm_squared[ i - 6 ] = sample_average[ i ]
                    write_line_success = write_data_to_file( UID, timestamp, decimal_hour, batch_number,
                        visible_irradiance_uW_per_cm_squared, nir_irradiance_uW_per_cm_squared,
                        surface_temperature_C, air_temperature_C, relative_humidity_percent,
                        main_battery_voltage, onboard_neopixel )
                    gc.collect()
                    mirror_data_over_usb(
                        UID, timestamp, decimal_hour, batch_number,
                        averaged_visible_irradiance_uW_per_cm_squared, averaged_nir_irradiance_uW_per_cm_squared,
                        surface_temperature_C, air_temperature_C, relative_humidity_percent)
                    time.sleep( 1 )
                    display_mode = SHOW_BATCH_NUMBER
                    last_instrument_mode = HOLD
                    instrument_mode = HOLD
                else:
                    print( "print 'no sd' to display" )

            # Wait loop
            gc.collect()
            wait_memory_free = gc.mem_free()
            if main_battery_voltage < LOW_BATTERY_VOLTAGE:
                low_battery_voltage_notification( four_digit_display )
            #print("at wait loop, memory free {} B".format( wait_memory_free ))
            remaining_wait_time = desired_sample_interval_s - (time.monotonic() - last_sample_time_s)
            while remaining_wait_time > 0 and exit_wait == 0:
                # check for keys pressed
                orange_key_pressed, yellow_key_pressed, green_key_pressed, blue_key_pressed = check_neokey( neokey1x4 )
                # turn key light off to acknowledge press detected
                if neokey1x4:
                    if orange_key_pressed:
                        neokey1x4.pixels[0] = HEX_OFF
                    else:
                        neokey1x4.pixels[0] = HEX_ORANGE
                    if yellow_key_pressed:
                        neokey1x4.pixels[1] = HEX_OFF
                    else:
                        neokey1x4.pixels[1] = HEX_YELLOW
                    if green_key_pressed:
                        neokey1x4.pixels[2] = HEX_OFF
                    else:
                        neokey1x4.pixels[2] = HEX_GREEN
                    if blue_key_pressed:
                        neokey1x4.pixels[3] = HEX_OFF
                    else:
                        neokey1x4.pixels[3] = HEX_BLUE


                if orange_key_pressed and not last_orange_key_pressed:
                    print( "check to see if orange key is being held down for a full 2 seconds" )
                    loop_count = 0
                    while check_neokey( neokey1x4 )[ 0 ] and loop_count < 30:
                        loop_count += 1
                        time.sleep( 0.01 )
                    #print( "loop count = {}". format ( loop_count ))
                    if loop_count > 25:
                        display_mode = SHOW_BATCH_NUMBER
                        print( "go into date and time set mode" ) # ------------ date and time set mode -------------------------------------
                        neokey1x4.pixels[0] = 0xFF5500
                        gc.collect()
                        print( "memory free before going into date time update = {}".format( gc.mem_free() ))
                        updated_timestamp = set_date_time( rtc, timestamp, four_digit_display, neokey1x4 ) #check and update the date and time
                        print( "update_timestamp == {}".format( updated_timestamp ))
                        try:
                            rtc.datetime = updated_timestamp
                        except ValueError:
                            print( "Error: unable to update hardware real time clock" )
                        while neokey1x4[ 0 ]:
                            print( "waiting for orange key release" )
                            time.sleep( 0.01 )
                        while not neokey1x4[ 0 ]:
                            timestamp = rtc.datetime
                            four_digit_display.print("{:02}:{:02}".format( timestamp.tm_min, timestamp.tm_sec ))
                            print( "waiting for orange key press" )
                            time.sleep( 0.25 )
                    display_mode = ( display_mode + 1 ) % number_of_display_modes
                    if instrument_mode == HOLD:
                        instrument_mode = LIVE
                last_orange_key_pressed = orange_key_pressed

                if yellow_key_pressed and not last_yellow_key_pressed:
                    if instrument_mode != HOLD:
                        instrument_mode = HOLD
                    last_instrument_mode = RECORD
                last_yellow_key_pressed = yellow_key_pressed

                if blue_key_pressed and not last_blue_key_pressed:
                    if instrument_mode != RECORD:
                        instrument_mode = RECORD
                    else:
                        instrument_mode = LIVE
                last_blue_key_pressed = blue_key_pressed

                if green_key_pressed and not last_green_key_pressed:
                    if instrument_mode != AVERAGE:
                        instrument_mode = AVERAGE
                last_green_key_pressed = green_key_pressed

                #print( display_mode )
                remaining_wait_time = desired_sample_interval_s - (time.monotonic() - last_sample_time_s)
                #print( "remaining_wait_time == {:}".format(remaining_wait_time))
                time.sleep( 0.1 )
    finally:   # clean up the busses when ctrl-c'ing out of the loop
        sdcard.deinit()
        print( "sd card deinitialized" )
        i2c_bus.deinit()
        print( "i2c_bus deinitialized" )

    print( "STELLA done" )

# define functions

gc.collect()

def mirror_data_over_usb(
    UID, timestamp, decimal_hour, batch_number,
    visible_irradiance_uW_per_cm_squared, nir_irradiance_uW_per_cm_squared,
    surface_temperature_C, air_temperature_C, relative_humidity_percent
    ):
    if True: #False:
        gc.collect()
        #print( "memory free immediately before mirror data == {} kB.".format( gc.mem_free()/1000))
        print( "mirror_data", end=", " )
        print( "UID, {}".format( UID ), end=", ")
        print( "batch, {}".format( batch_number ), end=", ")
        print( "year, {}".format( timestamp.tm_year ), end=", ")
        print( "month, {}".format( timestamp.tm_mon ), end=", ")
        print( "day, {}".format( timestamp.tm_mday ), end=", ")
        print( "hour, {}".format( timestamp.tm_hour ), end=", ")
        print( "min, {}".format( timestamp.tm_min ), end=", ")
        print( "sec, {}".format( timestamp.tm_sec ), end=", ")
        print( "dec_hour, {}".format( decimal_hour ), end=", ")
        print( "v450, {}".format( visible_irradiance_uW_per_cm_squared[ 0 ]), end=", ")
        print( "b500, {}".format( visible_irradiance_uW_per_cm_squared[ 1 ]), end=", ")
        print( "g550, {}".format( visible_irradiance_uW_per_cm_squared[ 2 ]), end=", ")
        print( "y570, {}".format( visible_irradiance_uW_per_cm_squared[ 3 ]), end=", ")
        print( "o600, {}".format( visible_irradiance_uW_per_cm_squared[ 4 ]), end=", ")
        print( "r650, {}".format( visible_irradiance_uW_per_cm_squared[ 5 ]), end=", ")
        print( "610, {}".format( nir_irradiance_uW_per_cm_squared[ 0 ]), end=", ")
        print( "680, {}".format( nir_irradiance_uW_per_cm_squared[ 1 ]), end=", ")
        print( "730, {}".format( nir_irradiance_uW_per_cm_squared[ 2 ]), end=", ")
        print( "760, {}".format( nir_irradiance_uW_per_cm_squared[ 3 ]), end=", ")
        print( "810, {}".format( nir_irradiance_uW_per_cm_squared[ 4 ]), end=", ")
        print( "860, {}".format( nir_irradiance_uW_per_cm_squared[ 5 ]), end=", ")
        print( "surface_temp, {}".format( surface_temperature_C ), end=", ")
        print( "air_temp, {}".format( air_temperature_C ), end=", ")
        print( "rel_humidity, {}".format( relative_humidity_percent ))
    return True

def check_neokey( neokey_object ):
    gc.collect()
    keys = [ False, False, False, False ]
    if neokey_object:
        try:
            keys[ 0 ] = neokey_object[ 0 ]
            keys[ 1 ] = neokey_object[ 1 ]
            keys[ 2 ] = neokey_object[ 2 ]
            keys[ 3 ] = neokey_object[ 3 ]
        except ValueError:
            keys = [ False, False, False, False ]
    return keys

'''
def read_TIR_sensor( TIR_sensor ):
    if TIR_sensor:
        decimal_places = 1
        try:
            surface_temperature_C = TIR_sensor.object_temperature
            surface_temperature_C = round( surface_temperature_C, decimal_places )
            return surface_temperature_C
        except ValueError as err:
            print( "Error: thermal infrared sensor fail: {:}".format(err) )
            return -273
    else:
        return -273
'''

def read_environmental_sensors( temp_hum_SHT40 ):
    gc.collect()
    if temp_hum_SHT40:
        air_temperature_C, relative_humidity_percent = temp_hum_SHT40.measurements
        air_temperature_C = round( air_temperature_C, 1 )
        relative_humidity_percent = round( relative_humidity_percent, 1 )
    else:
        air_temperature_C = 0
        relative_humidity_percent = 0
    return air_temperature_C, relative_humidity_percent

def read_spectrum( spectral_sensor ):
    gc.collect()
    decimal_places = 1
    # returns readings in a tuple, six channels of data
    # Calibration information from AS7262 datasheet:
    # Each channel is tested with GAIN = 16x, Integration Time (INT_T) = 166ms and VDD = VDD1 = VDD2 = 3.3V, TAMB=25°C.
    # The accuracy of the channel counts/μW/cm2 is ±12%.
    # Sensor Mode 2 is a continuous conversion of light into data on all 6 channels
    # 450nm, 500nm, 550nm, 570nm, 600nm and 650nm
    # sensor.violet returns the calibrated floating point value in the violet channel.
    # sensor.raw_violet returns the uncalibrated decimal count value in the violet channel.
    # that syntax is the same for each of the 6 channels

    # NOTE: the library was written for the visible sensor, as7262.
    # The library works exactly the same for the near infrared sensor, as7263,
    # but the band names are called as if the sensor were the visible one.
    try:
        while not spectral_sensor.data_ready:
            time.sleep(0.01)
        initial_read_time_s = time.monotonic()
        violet_calibrated = int( round( spectral_sensor.violet, decimal_places ))
        if violet_calibrated > MAX_VALUE:
            violet_calibrated = MAX_VALUE
        blue_calibrated = int( round( spectral_sensor.blue, decimal_places ))
        if blue_calibrated > MAX_VALUE:
            blue_calibrated = MAX_VALUE
        green_calibrated = int( round( spectral_sensor.green, decimal_places ))
        if green_calibrated > MAX_VALUE:
            green_calibrated = MAX_VALUE
        yellow_calibrated = int( round( spectral_sensor.yellow, decimal_places ))
        if yellow_calibrated > MAX_VALUE:
            yellow_calibrated = MAX_VALUE
        orange_calibrated = int( round( spectral_sensor.orange, decimal_places ))
        if orange_calibrated > MAX_VALUE:
            orange_calibrated = MAX_VALUE
        red_calibrated = int( round( spectral_sensor.red, decimal_places ))
        if red_calibrated > MAX_VALUE:
            red_calibrated = MAX_VALUE
        final_read_time_s = time.monotonic()
        read_time_s = final_read_time_s - initial_read_time_s
        #print( "Sensor read time s = {}".format( read_time_s ))
        spectral_tuple = ( violet_calibrated, blue_calibrated, green_calibrated, yellow_calibrated, orange_calibrated, red_calibrated )
        if HUGE_VALUE:
            spectral_tuple = ( HUGE_VALUE, HUGE_VALUE, HUGE_VALUE, HUGE_VALUE, HUGE_VALUE, HUGE_VALUE)
        return spectral_tuple
    except ValueError as err:
        print( "Error: spectral sensor fail: {:}".format(err) )
        return ( 0, 0, 0, 0, 0, 0)

def low_battery_voltage_notification( four_digit_display ):
    four_digit_display.print( "LO B" )
    print( "main battery low, plug in to recharge" )

def calculate_and_write_standard_deviations( UID, batch_number, filename ):
    try:
        with open( filename, "r") as sf:
            sf.readline() #throw away the header line
            line = sf.readline().split(',')
            number_of_samples = int( line[ 2 ] )
            #print( "number of samples == {}".format( number_of_samples ))
            number_of_channels = 12
            sf.close()
            stdev_list = []
            stdev_list=[str(UID), str(batch_number),"standard_deviation"]
            for n in range( 0, number_of_channels ):
                #print( "n == {}".format( n ))
                with open( filename, "r") as sf:
                    dataline=[]
                    sf.readline() #throw away the header line
                    for i in range( 0, number_of_samples ):
                        #print( "i == {}".format( i ))
                        dataline.append( int((sf.readline().split(','))[ n + 3 ]))
                        #print( dataline )
                    #print( dataline )
                    stdev_list.append( "{}".format( ulab.numpy.std( dataline )))
                sf.close()
            print( stdev_list )
            print( "STANDARD DEVIATION SUCCESS!" )

        with open( filename, "a") as sf:
            #sf.write("\n")
            for i,value in enumerate( stdev_list ):
                sf.write( "{},".format( value ))
            sf.close()
        return True
    except (MemoryError, IndexError) as err:
        print( "{}: could not complete standard deviation operation".format(err) )
        time.sleep(2)
        return False

def write_tail_of_sampling_file( UID, batch_number, iso8601_utc,
    onboard_neopixel, sampling_filename, averaging_number_of_samples,
    sample_average, VIS_sensor_temperature, NIR_sensor_temperature, surface_temperature_C,
    air_temperature_C, relative_humidity_percent
    ):
    gc.collect()
    if onboard_neopixel:
        set_neopixel( onboard_neopixel, NEO_RED )
    try:
        with open( sampling_filename, "a" ) as sf:
            sf.write( str( UID ))
            sf.write(",")
            sf.write( str( batch_number ))
            sf.write( ", average, " )
            for i,v in enumerate( sample_average ):
                sf.write( str(v) )
                #if i < len(sample_average)-1: # no trailing ','
                sf.write(",")
            #sf.write( "\n" )
            sf.write( " {}, {}, {}, {}, {}, {}\n".format(
                VIS_sensor_temperature, NIR_sensor_temperature, surface_temperature_C,
                air_temperature_C, relative_humidity_percent, iso8601_utc
                ))
        if onboard_neopixel:
            set_neopixel( onboard_neopixel, NEO_OFF )
        return True
    except OSError:
        return False

def write_sampling_line_to_file( UID, batch_number, iso8601_utc, onboard_neopixel, sampling_filename, sampling_index, visible_spectrum_uW_per_cm_squared, nir_spectrum_uW_per_cm_squared ):
    gc.collect()
    with open( sampling_filename, "r" ) as sf:
        sf.flush()
        sf.close()

    if onboard_neopixel:
        set_neopixel( onboard_neopixel, NEO_RED )
    try:
        with open( sampling_filename, "a" ) as sf:
            sf.write( str( UID ))
            sf.write(",")
            sf.write( str( batch_number ))
            sf.write( ", " )
            sf.write( str( sampling_index ))
            sf.write( ", " )
            for i,v in enumerate(visible_spectrum_uW_per_cm_squared):
                sf.write( str(v) )
                sf.write(",")
            for i,v in enumerate(nir_spectrum_uW_per_cm_squared):
                sf.write( str(v) )
                #if i < len(nir_spectrum_uW_per_cm_squared)-1: # no trailing ','
                sf.write(",")
            sf.write( "0, 0, 0, 0, 0, ")
            sf.write( str( iso8601_utc ))
            sf.write( "\n" )
            sf.flush()
            sf.close()
        if onboard_neopixel:
            set_neopixel( onboard_neopixel, NEO_OFF )
        return True
    except OSError:
        return False

def get_spectral_sensor_temperatures(VIS_sensor, NIR_sensor):
    vis_temp_C = round( VIS_sensor.temperature, 1 )
    nir_temp_C = round( NIR_sensor.temperature, 1 )
    return vis_temp_C, nir_temp_C

def create_sampling_file( UID, batch_number, timestamp, decimal_hour, values ):
    gc.collect()
    sampling_filename = "/sd/{}_{:04}{:02}{:02}_{}.csv".format( UID, timestamp.tm_year, timestamp.tm_mon, timestamp.tm_mday, batch_number)
    try:
        with open( sampling_filename, "w" ) as sf:
            sf.write( "UID, batch, sample_number, V450_irradiance, B500_irradiance, G550_irradiance, Y570_irradiance, O600_irradiance, "
                "R650_irradiance, nir610_irradiance, nir680_irradiance, nir730_irradiance, nir760_irradiance, "
                "nir810_irradiance, nir860_irradiance,  VIS_sensor_temperature,  NIR_sensor_temperature, "
                "surface_temperature, air_temperature, relative_humidity, timestamp\n" )
            sf.close()
        return sampling_filename
    except OSError:
        return False

def initialize_data_file():
    need_header = False
    try:
        os.stat( DATA_FILE ) # fail if data.csv file is not already there
        #raise OSError # uncomment to force header everytime
        return False
    except OSError:
        # setup the header for first time
        need_header = True
        #print( "need header = True")
        if need_header:
            preheader_mem_free = gc.mem_free()
            print( "preheader mem free = {}".format(preheader_mem_free))
            gc.collect()
            try:
                with open( DATA_FILE, "w" ) as f:
                    header = (
                    "UID, batch_number, weekday, timestamp, decimal_hour, "
                    "surface_temperature_C, surface_temperature_error_C, "
                    "air_temperature_C, air_temperature_error_C, "
                    "relative_humidity_percent, relative_humidity_error_percent, "
                    "V450_wavelength_nm, V450_wavelength_error_nm, V450_irradiance_uW_per_cm_squared, V450_irradiance_error_uW_per_cm_squared, "
                    "B500_wavelength_nm, B500_wavelength_error_nm, B500_irradiance_uW_per_cm_squared, B500_irradiance_error_uW_per_cm_squared, "
                    "G550_wavelength_nm, G550_wavelength_error_nm, G550_irradiance_uW_per_cm_squared, G550_irradiance_error_uW_per_cm_squared, "
                    "Y570_wavelength_nm, Y570_wavelength_error_nm, Y570_irradiance_uW_per_cm_squared, Y570_irradiance_error_uW_per_cm_squared, "
                    "O600_wavelength_nm, O600_wavelength_error_nm, O600_irradiance_uW_per_cm_squared, O600_irradiance_error_uW_per_cm_squared, "
                    "R650_wavelength_nm, R650_wavelength_error_nm, R650_irradiance_uW_per_cm_squared, R650_irradiance_error_uW_per_cm_squared, "
                    "nir610_wavelength_nm, nir610_wavelength_error_nm, nir610_irradiance_uW_per_cm_squared, nir610_irradiance_error_uW_per_cm_squared, "
                    "nir680_wavelength_nm, nir680_wavelength_error_nm, nir680_irradiance_uW_per_cm_squared, nir680_irradiance_error_uW_per_cm_squared, "
                    "nir730_wavelength_nm, nir730_wavelength_error_nm, nir730_irradiance_uW_per_cm_squared, nir730_irradiance_error_uW_per_cm_squared, "
                    "nir760_wavelength_nm, nir760_wavelength_error_nm, nir760_irradiance_uW_per_cm_squared, nir760_irradiance_error_uW_per_cm_squared, "
                    "nir810_wavelength_nm, nir810_wavelength_error_nm, nir810_irradiance_uW_per_cm_squared, nir810_irradiance_error_uW_per_cm_squared, "
                    "nir860_wavelength_nm, nir860_wavelength_error_nm, nir860_irradiance_uW_per_cm_squared, nir860_irradiance_error_uW_per_cm_squared, "
                    "battery_voltage"
                    "\n"
                        )

                    f.write( header )
            except OSError:
                pass
            postheader_mem_free = gc.mem_free()
            print( "header mem usage = {}".format(postheader_mem_free - preheader_mem_free))
        return True

def write_data_to_file( UID, timestamp, decimal_hour, batch_number,
                visible_irradiance_uW_per_cm_squared, nir_irradiance_uW_per_cm_squared,
                surface_temperature_C, air_temperature_C, relative_humidity_percent,
                main_battery_voltage, onboard_neopixel ):
    # we let operations fail if the sdcard didn't initialize
    try:
        os.stat( DATA_FILE ) # fail if data.cav file is not already there
        #raise OSError # uncomment to force header everytime
    except OSError:
        # setup the header for first time
        if onboard_neopixel:
            set_neopixel( onboard_neopixel, NEO_RED )
        return False
    try:
        # with sys.stdout as f: # uncomment to write to console (comment "with open")
        with open( DATA_FILE, "a" ) as f: # open seems about 0.016 secs.
            if onboard_neopixel:
                set_neopixel( onboard_neopixel, NEO_RED )

            iso8601_utc = "{:04}{:02}{:02}T{:02}{:02}{:02}Z".format( timestamp.tm_year,
            timestamp.tm_mon, timestamp.tm_mday, timestamp.tm_hour, timestamp.tm_min,
            timestamp.tm_sec
            )
            weekday = timestamp_to_weekday( timestamp )
            decimal_hour = timestamp_to_decimal_hour( timestamp )
            f.write( "{}, {}, {}, {}, {}, ".format( UID, batch_number, weekday, iso8601_utc, decimal_hour ))
            f.write( "{}, 1.0, ".format( surface_temperature_C ))
            f.write( "{}, 0.3, ".format( air_temperature_C ))
            f.write( "{}, 1.8, ".format( relative_humidity_percent ))
            for i,band in enumerate( VIS_BANDS ):
                f.write( str(VIS_BANDS[ i ]) )
                f.write( ", ")
                f.write( "5, " )
                f.write( str(visible_irradiance_uW_per_cm_squared[ i ] ))
                f.write( ", ")
                f.write( str(visible_irradiance_uW_per_cm_squared[ i ] * 12/100) )
                f.write( ", ")
            for i,band in enumerate( NIR_BANDS ):
                f.write( str(NIR_BANDS[ i ] ))
                f.write( ", ")
                f.write( "5, " )
                f.write( str(nir_irradiance_uW_per_cm_squared[ i ] ))
                f.write( ", ")
                f.write( str( nir_irradiance_uW_per_cm_squared[ i ] * 12/100 ))
                f.write( ", ")
            f.write( "{}".format( main_battery_voltage ) )
            f.write("\n")
        if onboard_neopixel:
            time.sleep( 0.05 )
            set_neopixel( onboard_neopixel, NEO_OFF )
        return True
    except OSError as err:
        # TBD: maybe show something on the display like sd_full? this will "Error" every sample pass
        # "[Errno 30] Read-only filesystem" probably means no sd_card
        print( "Error: sd card fail: {:} ".format(err) )
        if onboard_neopixel:
            set_neopixel( onboard_neopixel, NEO_RED ) #  ant ON to show error, likely no SD card present, or SD card full.
        return False

def timestamp_to_weekday( timestamp ):
    try:
        clock_days = ( "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" )
        weekday = clock_days[ timestamp.tm_wday ]
        return weekday
    except ValueError as err:
        print( "Error: real time hardware_clock fail: {:}".format(err) )
        return "Error"

def timestamp_to_decimal_hour( timestamp ):
    try:
        decimal_hour = timestamp.tm_hour + timestamp.tm_min/60.0 + timestamp.tm_sec/3600.0
        return decimal_hour
    except ValueError as err:
        print( "Error: invalid timestamp: {:}".format(err) )
        return False

def initialize_four_digit_display( bus_mux, new_address ):
    try:
        display = segments.Seg7x4( bus_mux, address = new_address )
        print("display succeeded")
        display.fill(0)
    except ValueError:
        print( "4 digit 7 segment display not found" )
        display = False
    return display


def initialize_neokey1x4( i2c_bus ):
    try:
        neokey = NeoKey1x4(i2c_bus, addr=0x30)
    except ValueError:
        print( "neokey 1x4 module not found" )
        neokey = False
    return neokey

def initialize_spectral_sensor( bus_mux, mux_channel ):
    try:
        spectral_sensor = AS726x_I2C( bus_mux[ mux_channel ])
        #print( "\nSet spectral sensor at mux channel {} to 16x gain and 166 ms integration time, because those are the values used at the factory during the sensor calibration.".format( mux_channel ))
        spectral_sensor.conversion_mode = spectral_sensor.MODE_2
        #print( "Default Gain = {}".format( spectral_sensor.gain ))
        spectral_sensor.gain = 16
        #print( "MUX channel {}: Gain = {}".format( mux_channel, spectral_sensor.gain ))
        #print( "Default Integration Time = {} ms".format( spectral_sensor.integration_time ))
        spectral_sensor.integration_time = 166 # from 1 to 714 = 255 * 2.8ms ms
        print( "Spectral Sensor on MUX channel {}: Gain = {}x, Integration Time = {} ms".format( mux_channel, spectral_sensor.gain, spectral_sensor.integration_time ))
        return spectral_sensor
    except ValueError:
        print( "spectral_sensor at channel {} not found".format( mux_channel ))
        return False

def initialize_tca_i2c_multiplexer( i2c_bus ):
    try:
        tca = adafruit_tca9548a.TCA9548A( i2c_bus )
    except ValueError:
        print( "i2c bus multiplexer not available" )
        tca = False
    return tca

def initialize_SHT40( i2c_bus ):
    try:
        sht = adafruit_sht4x.SHT4x( i2c_bus )
    except ValueError:
        print( "sht40 temp_humid could not be found" )
        sht = False
    try:
        #print( "Found SHT40 with serial number", hex( sht.serial_number ))
        sht.mode = adafruit_sht4x.Mode.NOHEAT_HIGHPRECISION
        # Can also set the mode to enable heater
        # sht.mode = adafruit_sht4x.Mode.LOWHEAT_100MS
        print("SHT40 air temperature and humidity mode: {}".format( adafruit_sht4x.Mode.string[ sht.mode ]))
    except( OSError, AttributeError ):
        print( "SHT40 sensor serial number lookup failed" )
    return sht

def check_STELLA_config():
    try:
        with open( "/sd/STELLA_config.txt", "r" ) as s:
            try:
                header = s.readline()
                #print( sampling_header )
                values = s.readline()
                #print(sampling_values)
                values= values.rstrip('\n')
                values = values.split(',')
                #print(line)
                desired_sample_interval_s = float( values[ 0 ])
                averaging_number_of_samples = int( values[ 1 ])
                averaging_count_down = int( values[ 2 ])
                count_down_step_s = float( values[ 3 ])

            # TBD: catch error when /sd doesn't exist
            except ValueError:
                pass
    except OSError:
            print( "STELLA_config.txt file not found" )
            try:
                with open( "/sd/STELLA_config.txt", "w" ) as s:
                    s.write( "desired_sample_interval_s, averaging_number_of_samples, averaging_count_down, count_down_step_s\n" )
                    s.write( "0.75, 20, 3, 0.75\n" )
                print("created new STELLA_config.txt file")
            except OSError:
                pass
            desired_sample_interval_s, averaging_number_of_samples, averaging_count_down, count_down_step_s = 0.75, 21, 3, 0.75
    return desired_sample_interval_s, averaging_number_of_samples, averaging_count_down, count_down_step_s

def update_batch( timestamp ):
    gc.collect()
    datestamp = "{:04}{:02}{:02}".format( timestamp.tm_year, timestamp.tm_mon, timestamp.tm_mday)
    try:
        with open( "/sd/batch.txt", "r" ) as b:
            try:
                previous_batchfile_string = b.readline()
                previous_datestamp = previous_batchfile_string[ 0:8 ]
                previous_batch_number = int( previous_batchfile_string[ 8: ])
            # TBD: catch error when /sd doesn't exist
            except ValueError:
                previous_batch_number = 0
                # corrupted data in batch number file, setting batch to 0
            if datestamp == previous_datestamp:
                # this is the same day, so increment the batch number
                batch_number = previous_batch_number + 1
            else:
                # this is a different day, so start the batch number at 0
                batch_number = 0
    except OSError:
            print( "batch.txt file not found" )
            batch_number = 0
    batch_string = ( "{:03}".format( batch_number ))
    batch_file_string = datestamp + batch_string
    try:
        with open( "/sd/batch.txt", "w" ) as b:
            b.write( batch_file_string )
        # TBD: catch error when /sd doesn't exist
    except OSError as err:
        print("Error: writing batch.txt {:}".format(err) )
        pass
    batch_string = ( "{:}".format( batch_number ))
    return batch_string

def initialize_real_time_clock( i2c_bus ):
    try:
        rtc = adafruit_pcf8523.PCF8523( i2c_bus )
        clock_battery_OK = not rtc.battery_low
        if clock_battery_OK:
            print( "clock battery is OK." )
        else:
            print( "clock battery is low. Replace the clock battery." )
    except ( NameError, ValueError ):
        print( "real time clock failed to initialize" )
        rtc = False
        clock_battery_OK = False
    return ( rtc, clock_battery_OK )

def initialize_i2c_bus( SCL_pin, SDA_pin ):
    try:
        i2c_bus = busio.I2C( SCL_pin, SDA_pin )
    except ValueError as err:
        print( "i2c bus failed to initialize '{}'".format( err ))
        i2c_bus = False
    return i2c_bus

def set_neopixel( neopixel_object, color ):
    if neopixel_object:
        if color:
            neopixel_object[ 0 ] = color
        else:
            neopixel_object[ 0 ] = ( 0, 0, 0 )
        neopixel_object.show()
    else:
        return False

def initialize_neopixel( pin_number, number_of_pixels ):
    try:
        neopixel_object = neopixel.NeoPixel( pin_number, number_of_pixels, brightness=1, auto_write=False, pixel_order=neopixel.GRB )
    except ValueError:
        print( "neopixel initialization failed." )
        neopixel_object = False
    return neopixel_object

def check_battery_voltage( battery_monitor ):
    if battery_monitor:
        battery_voltage = round(( battery_monitor.value * 3.3) / 65536 * 2, 2 )
        return battery_voltage
    else:
        return 0

def initialize_discrete_LED( pin_number ):
    try:
        discrete_LED = digitalio.DigitalInOut( pin_number )
        discrete_LED.direction = digitalio.Direction.OUTPUT
    except ValueError:
        print( "discrete LED initialization failed." )
        discrete_LED = False
    return discrete_LED

def initialize_AnalogIn( pin_number ):
    try:
        analog_signal = AnalogIn( pin_number )
    except ValueError:
        print( "analog input initialization failed." )
        analog_signal = False
    return analog_signal

gc.collect()
#print("memory free after loading function definitions into RAM = {} bytes".format( gc.mem_free() ))

main()
